@@ -1,7 +1,6 @@
 #!/usr/bin/env python3
 
 import asyncio
-from collections import OrderedDict
 from functools import partial
 from itertools import chain
 from threading import Barrier, Thread
@@ -18,6 +17,8 @@ from .utils import aiter, isdict, islist
 class Node(Component):
     """Processes messages."""
 
+    globals = Box()
+
     def configure(self, process=None, inbox=None, ifilters=None, ofilters=None, outbox=None,
                   conditions=None, blocking=False, scale=None, **settings):
         if callable(process):
@@ -33,6 +34,12 @@ class Node(Component):
         self.blocking = self.__class__ is ThreadedNode or blocking
         self.scale = int(scale or 1)
         self.layers = list()
+        self.locals = Box()
+
+        # Used by ProcessorWrapper to control behavior of input feed.
+        self.wait_for_pending = True
+
+        self.streams = dict()
         return settings
 
     def __str__(self):
@@ -50,6 +57,14 @@ class Node(Component):
 
     _dependent_statuses = set()
 
+    async def test(self, messages):
+        self.debug(f'~~~~~~ begin {messages}')
+        async for c, m in messages:
+            self.debug(f'~~~~~~ message from {messages} {c} {m}')
+            yield c, m
+            self.debug(f'~~~~~~ yielded from {messages} {c} {m}')
+        self.debug(f'~~~~~~ end {messages}')
+
     def start(self, *args):
         coroutines = [super().start(*args)]
         self.layers = ([self.inbox] +
@@ -91,16 +106,32 @@ class Node(Component):
 
     async def _run(self):
         if not self.running:
+            self.debug('*** _run exiting')
             return
 
+        # for layer in self.layers:
+        #     if isinstance(layer, Component) and hasattr(layer, 'ready'):
+        #         self.debug(f'Waiting on layer {layer} to be ready')
+        #         await layer.ready.wait()
+        #         self.debug(f'Layer {layer} is ready')
         stack = self.layers[0]
         for layer in self.layers[1:]:
             stack = layer(stack)
+            # stack = layer(self.test(stack))
+        # async for result in self.test(stack):
         async for result in stack:
             if not self.running:
                 break
 
     def spawn_processor(self):
+        # self._processor = self.get_processor()
+        # if self.blocking:
+        #     raise NotImplementedError(f'Node.blocking in {self}')
+        #     # self._processor = ThreadedProcessor(self._processor, scale=self.scale)
+
+        # if self.scale > 1:
+        #     self._processor = MultiProcessor(self)
+
         self._processor = ProcessorWrapper(self)
         return [self._pre_processor, self._processor, self._post_processor]
 
@@ -108,6 +139,7 @@ class Node(Component):
         """Generate an event on the first message reaching the processor."""
         first_message = True
         async for channel, message in messages:
+            # self.debug(f'*** pre-proc {channel}: {message}')
             if not self.running:
                 break
 
@@ -115,16 +147,37 @@ class Node(Component):
                 self.status = 'processing-started'
                 first_message = False
             yield (channel, message)
+            # self.debug('*** pre-proc loop')
+        # self.debug('*** pre-proc end')
 
     async def _post_processor(self, messages):
         """Generate an event on the last message leaving the processor."""
         async for channel, message in messages:
+            # self.debug(f'*** post-proc {channel}: {message}')
             if not self.running:
                 break
 
             yield (channel, message)
+            # self.debug('*** post-proc loop')
+        # self.debug('*** post-proc end')
         self.status = 'processing-finished'
 
+    # def get_processor(self):
+    #     if isinstance(self, WithRetry):
+    #         self._retryable = None
+    #         return self._retryable_processor
+    #
+    #     return self.processor
+
+    # async def _retryable_processor(self, messages, id_=None):
+    #     if self._retryable is None:
+    #         self._retryable = RetryableMessages(self, messages)
+    #     async for channel, message in self.processor(self._retryable, id_):
+    #         # self.debug(f'*** _retryable {channel}: {message}')
+    #         yield channel, message
+    #         # self.debug('*** _retryable loop')
+    #     # self.debug('*** _retryable end')
+
     async def processor(self, messages, id_=None):
         """Process and yield new (channel, message)."""
         async for channel, message in messages:
@@ -138,9 +191,9 @@ class Node(Component):
     def drop(self, message):  # TODO: implement message accounting and leak detection
         self._processor.drop(message)
 
-    async def retry(self, channel=None, message=None, key=None, wait=False):
+    async def retry(self, channel, message, wait=False):
         """Insert a message back into the input queue for retrying it."""
-        await self._processor.retry(channel, message, key, wait)
+        await self._processor.retry(channel, message, wait)
 
     def merged_settings(self, message, key=None, msgmap=None):
         """Return node settings, with values overridden from the message if present."""
@@ -189,37 +242,50 @@ class ProcessorWrapper:
         self.node = node
         self.running = True
         self.exc = None
-        self.backlog = OrderedDict()
+        # self.processor = self.node.processor
 
     async def infeed(self, messages):
         """Loop through all messages from the inbox and relay them to the input queue."""
+        self.node.debug('*** multi-processor infeed running')
         async for channel, message in messages:
+            # TODO: account for dropped messages
+            self.node.debug(f'*** multi-processor infeed rcv msg pending={self.pending} {message}')  # noqa: E501
             self.pending[f'{id(message)}-{message.meta.id}'] = message
+            self.node.debug(f'*** infeed got {message} from {channel}')
             await self.iqueue.put((channel, message))
-        if self.backlog:
-            self.node.debug('No more messages, processing backlog')
-            for channel, message in self.backlog.copy().values():
-                await self.rqueue.put((channel, message))
-        pending = None
-        while self.pending and self.node.running:
-            if pending != len(self.pending):
-                pending = len(self.pending)
-            await asyncio.sleep(0.1)
+            self.node.debug(f'*** infeed put {message} from {channel}')
+
+        if self.pending:
+            if self.node.wait_for_pending:
+                pending = None
+                self.node.debug(f'*** infeed exhausted pending={pending}')
+                while self.pending and self.node.running:
+                    if pending != len(self.pending):
+                        pending = len(self.pending)
+                        self.node.debug(f'*** multi-processor infeed waiting pending={pending}')
+                    await asyncio.sleep(0.1)
+            if self.pending:
+                plural = 's' if len(self.pending) != 1 else ''
+                self.node.warning(f'ProcessorWrapper exiting with {len(self.pending)} pending '
+                                  f'message{plural}')
+        self.node.debug('*** infeed stopping collectors')
         self.running = False
+        self.node.debug('*** infeed exiting')
 
     async def _input_iter(self):
         wait = self.iwait
         while self.node.running and self.running:
-            input_is_empty = True
+            queues_are_empty = True
             for queue in (self.rqueue, self.iqueue):
                 try:
                     yield queue.get_nowait()
 
-                    input_is_empty = False
+                    queues_are_empty = False
                 except asyncio.QueueEmpty:
+                    # self.node.debug('*** _input_iter empty')
                     pass
 
-            if input_is_empty:
+            if queues_are_empty:
                 await asyncio.sleep(wait)
                 if wait < 0.1:
                     wait *= 2
@@ -228,9 +294,14 @@ class ProcessorWrapper:
 
     async def collector(self, id_):
         """Launch an instance of the node processor, and relay its output to the output queue."""
+        self.node.debug(f'*** multi-processor collector-{id_} running processor={self.node.processor}')  # noqa: E501
         async for channel, message in self.node.processor(self._input_iter(), id_=id_):
+            self.node.debug(f'*** multi-processor collector-{id_} rcv msg {message}')
             await self.oqueue.put((channel, message))
+            self.node.debug(f'*** multi-processor collector-{id_} snt msg {message}')
+        self.node.debug(f'*** multi-processor collector-{id_} stopped')
         await self.oqueue.put((None, pipe.EOT))
+        self.node.debug(f'*** multi-processor collector-{id_} exited')
 
     async def outfeed(self, messages):
         """Launch all processor instances and return an iterator that feeds on the output queue."""
@@ -238,14 +309,19 @@ class ProcessorWrapper:
         self.rqueue = asyncio.Queue()                         # retry queue
         self.oqueue = asyncio.Queue(maxsize=1)                # output queue
         self.iwait = 0.001  # initial scaleback wait value for input queue
+        # self.finished = asyncio.Event()
+        # self.exhausted = False
         self.pending = dict()  # messages currently handled by node processor
+        # coroutines = [self.node.loop.create_task(self.infeed(messages))]
         coroutines = list()
         for i in range(self.node.scale):
+            # coroutines.append(self.node.loop.create_task(self.collector(i)))
             coroutines.append(asyncio.ensure_future(self.collector(i)))
         coroutines.append(asyncio.ensure_future(self.infeed(messages)))
         running = self.node.scale
         wait = self.iwait
 
+        self.node.debug(f'*** multi-processor outfeed started {running} collectors')
         while running or not self.oqueue.empty():
             if self.oqueue.empty():
                 # Raise any exception from collectors.
@@ -258,6 +334,7 @@ class ProcessorWrapper:
                         pass
 
                 if self.node.hasstatus('aborted'):
+                    self.node.debug('*** outfeed aborting')
                     break
 
                 await asyncio.sleep(wait)
@@ -268,52 +345,57 @@ class ProcessorWrapper:
             else:
                 wait = self.iwait
 
+            self.node.debug(f'*** multi-processor outfeed waiting for msg pending={len(self.pending)}')  # noqa: E501
             channel, message = await self.oqueue.get()
+            self.node.debug(f'*** multi-processor outfeed rcv msg pending={len(self.pending)} {message}')  # noqa: E501
             self.oqueue.task_done()
             if message is pipe.EOT:
                 running -= 1
+                self.node.debug(f'*** multi-processor outfeed collector ended left={running}')
             else:
                 self.pending.pop(f'{id(message)}-{message.meta.id}', None)
+                # if self.pending == 0 and self.exhausted:
+                #     self.node.debug('*** multi-processor outfeed finishing')
+                #     # self.finished.set()
                 yield channel, message
 
+            self.node.debug(f'*** multi-processor outfeed processed msg pending={len(self.pending)} {message}')  # noqa: E501
+
+        self.node.debug('*** outfeed stopped running')
         self.running = False
         for coroutine in coroutines:
             if self.node.hasstatus('aborted'):
+                self.node.debug(f'*** outfeed cancelling {coroutine}')
                 coroutine.cancel()
+            self.node.debug(f'*** outfeed awaiting {coroutine}')
             try:
                 await coroutine
             except asyncio.CancelledError:
                 pass
 
+        self.node.debug('*** outfeed all awaited')
+
     def drop(self, message, component=None):
         """Drop message and stop tracking it."""
         self.pending.pop(f'{id(message)}-{message.meta.id}', None)
         message.drop(component or self.node)
 
-    async def retry(self, channel=None, message=None, key=None, wait=False):
+    async def retry(self, channel, message, wait=False):
         """Insert a message back into the input queue for retrying it."""
         self.node.debug(f'Retrying message {message}')
-        if message:
-            self.pending[f'{id(message)}-{message.meta.id}'] = message
+        self.pending[f'{id(message)}-{message.meta.id}'] = message
 
         async def retry(channel, message, wait):
-            await asyncio.sleep(wait)  # tight loop mitigation
+            if wait:
+                await asyncio.sleep(wait)  # tight loop mitigation
             await self.rqueue.put((channel, message))
 
-        if wait is False:
+        if wait is None:
             wait = 0.000001
-        if wait is True:
-            if key is None:
-                raise ValueError('Missing key')
 
-            self.node.debug(f'Stashing message {message}')
-            self.backlog[key] = (channel, message)
-        else:
-            if key:
-                channel, message = self.backlog.pop(key)
-            # Schedule push to retry queue to free up caller.
-            self.node.debug(f'Requeuing message {message}')
-            asyncio.ensure_future(retry(channel, message, wait))
+        # Schedule push to retry queue to free up caller.
+        self.node.debug(f'Requeuing message {message}')
+        asyncio.ensure_future(retry(channel, message, wait))
 
     __call__ = outfeed
 
@@ -327,6 +409,7 @@ class Inbox(pipe.Manifold):
         self.debug('Ready to receive')
         for channel, _pipe in self._active_channels():
             if not self.running:
+                # self.debug('*** NOT running anymore')
                 return
 
             if _pipe is None:
@@ -341,12 +424,16 @@ class Inbox(pipe.Manifold):
 
             active_channels += 1
             try:
+                # self.debug(f'*** about to receive message from {channel}')
                 message = await self.try_while_running(
                         partial(_pipe.receive, wait=last_channel))
+                # self.debug(f'*** received message from {channel}: {message}')
             except asyncio.QueueEmpty:
+                # self.debug(f'*** no message to receive from {channel}')
                 continue
 
             except ComponentInterrupted:
+                # self.debug('*** receiving interrupted')
                 return
 
             except Exception:
@@ -374,10 +461,60 @@ class Inbox(pipe.Manifold):
                 else:
                     yield channel, _pipe
 
+    # async def Xreceiver(self):
+    #     self.debug('Receiving')
+    #     self.debug(f'Setting up channels {list(self.channels)}')
+    #     # __import__('pudb').set_trace()
+    #     # import sys; sys.exit()
+    #     feeder = deque()
+    #     events = Box(received=asyncio.Event())
+    #     # receivers = dict()
+    #     for channel, pipe in self._active_channels():
+    #         self.debug(f'Setting up channel {channel}')
+    #         events[channel] = asyncio.Event()
+    #         events[channel].set()
+    #         # receivers[channel] = asyncio.ensure_future(
+    #         #     self.channel_receiver(channel, feeder, events))
+    #         asyncio.ensure_future(self.channel_receiver(channel, feeder, events))
+    #
+    #     self.debug('Waiting for feeder')
+    #     while self.running and (len(feeder) or await events.received.wait()):
+    #         for channel in self.channels:
+    #             events[channel].clear()
+    #         events.received.clear()
+    #         channel, message = feeder.popleft()
+    #
+    #         self.debug(f'Got message from {channel}: {message}')
+    #         if message == pipe.EOT:
+    #             self.debug('Exhausted channel: %s' % channel)
+    #             pipe.stop()
+    #             # self.debug(f'* setting event for channel {channel}')
+    #             events[channel].set()
+    #             # receivers[channel].cancel()
+    #             try:
+    #                 await pipe.send(pipe.EOT)
+    #             except NotImplementedError:
+    #                 pass
+    #             # self.debug('-----------------------------')
+    #             if not list(self._active_channels()):
+    #                 self.debug('Stopping: no more active channels')
+    #                 self.stop()
+    #                 # TODO: cancel channel receiver coroutines
+    #                 break
+    #         else:
+    #             message.checkin(self)
+    #             yield channel, message
+    #
+    #         for channel in self.channels:
+    #             events[channel].set()
+    #
+    #     self.debug('Finished receiving')
+
     async def channel_receiver(self, channel, feeder, events):
         self.debug(f'Receiver for {channel} is ready')
         pipe = self.channels[channel]
         while pipe.running and self.running and await events[channel].wait():
+            # self.debug(f'* receiving from channel {channel}')
             events[channel].clear()
             try:
                 feeder.append((channel, await pipe.receive()))
@@ -389,6 +526,126 @@ class Inbox(pipe.Manifold):
             events.received.set()
         self.debug(f'Receiver for {channel} is exiting')
 
+    def X_active_channels(self):
+        channels = self.channels.copy()
+        while channels:
+            for channel, _pipe in channels.copy().items():
+                del channels[channel]
+                if not _pipe.running:
+                    self.debug(f'skipping stopped channel {channel}')
+                else:
+                    yield channel, _pipe
+        if channels:
+            pass  # TODO: channel clean-ups
+
+
+class RetryableMessages:
+    """Queue sitting between the inbox and the processor where retries can be inserted back."""
+
+    def __init__(self, node, messages):
+        self.node = node
+        self.messages = messages
+        self._queueiter = None
+
+    def __aiter__(self):
+        if self._queueiter is None:  # singleton queueiter, for when we're using MultiProcessor
+            self._queueiter = self.queueiter()
+        return self._queueiter
+
+    async def infeed(self):
+        """Feed the inbox to the queue."""
+
+        async def enqueue(channel, message):
+            self.forwarded.clear()
+            await self.queue.put((channel, message))
+            await self.node.try_while_running(partial(self.forwarded.wait))
+
+        # TODO: implement freeze detection - investigate where messages are stuck
+        try:
+            async for channel, self.newmsg in self.messages:
+                self.node.debug(f'*** retryable: feeding {self.newmsg}')
+                await enqueue(channel, self.newmsg)
+            self.node.debug(f'*** retryable: main infeed done (pending: {self.pending})')
+            while self.node.running and not self._retry_queue.empty():
+                channel, self.newmsg = await self._retry_queue.get()
+                self.node.debug(f'*** retryable: refeeding {self.newmsg}')
+                await enqueue(channel, self.newmsg)
+            self.node.debug(f'*** retryable: retry infeed done (pending: {self.pending})')
+            self.exhausted = True
+            if self.pending > 0:
+                self.node.debug(
+                        f'RetryableMessages waiting to exit: {self.pending} pending messages')
+                await self.node.try_while_running(partial(self.all_processing_done.wait))
+            self.node.debug('*** retryable: infeed ready to exit')
+        except ComponentInterrupted:
+            pass
+        finally:
+            await self.queue.put(pipe.EOT)
+        self.node.debug('*** retryable: infeed exited')
+
+    async def queueiter(self):
+        """Turn the queue into an iterator that the processor can consume."""
+        self.queue = asyncio.Queue()
+        self._retry_queue = asyncio.Queue()
+        self.forwarded = asyncio.Event()
+        self.all_processing_done = asyncio.Event()
+        self.exhausted = False
+        self.newmsg = None
+        self.pending = 0  # number of messages currently handled by node processor
+        infeed = self.node.loop.create_task(self.infeed())
+        self.node.debug('*** retryable: infeed created')
+
+        # while self.node.running:
+        #     try:
+        #         item = await self.node.try_while_running(self.queue.get)
+        #         if item == pipe.EOT:
+        #             self.queue.task_done()  # for EOT
+        #             break
+        #
+        #         else:
+        #             channel, message = item
+        #     except ComponentInterrupted:
+        #         pass
+        #     else:
+        #         # self.node.debug(f'*** retryable: forwarding {message.type} msg {message.meta.id}')  # noqa: E501
+        #         self.queue.task_done()
+        #         if message is self.newmsg:
+        #             self.forwarded.set()
+        #         self.pending += 1
+        #         yield channel, message
+
+        self.node.debug('*** retryable: consuming queue')
+        # messages = aiter(self.queue.get, pipe.EOT)
+        # try:
+        # async for channel, message in messages:
+        async for channel, message in aiter(self.queue.get, pipe.EOT):
+            self.node.debug(f'*** retryable: forwarding {message}')
+            self.queue.task_done()
+            if message is self.newmsg:
+                self.forwarded.set()
+            self.pending += 1
+            yield channel, message
+            self.node.debug(f'*** retryable: forwarded (pending: {self.pending}) {message}')  # noqa: E501
+
+        self.node.debug('*** retryable: queueiter exiting')
+        self.queue.task_done()
+        await infeed
+        self.node.debug('*** retryable: queueiter exited')
+
+    def processing_done(self):
+        """Signal that a message is not in the processor anymore."""
+        self.pending -= 1
+        if self.exhausted:
+            self.all_processing_done.set()
+
+    async def retry(self, channel, message, now=True, wait=0.001, id_=None):
+        """Insert a message back into the queue for retrying it."""
+        self.node.debug(f'*** retryable: retrying {message}')
+        await asyncio.sleep(wait)  # tight loop mitigation
+        queue = self.queue if self.exhausted else self._retry_queue
+        await queue.put((channel, message))
+        self.processing_done()
+
 
 class Outbox(pipe.Manifold):
     def __call__(self, messages):
@@ -401,7 +658,9 @@ class Outbox(pipe.Manifold):
                 yield channel, message
 
         except Exception:
+            self.error(f'*** outbox {self.id} caught an exception')
             self.status = 'aborted'
+            # self.exception('Node failure')
             raise
 
         else:
@@ -412,10 +671,12 @@ class Outbox(pipe.Manifold):
     async def sender(self, messages):
         async for channel, message in messages:
             channel = channel or 'default'
+            self.debug(f'*** got msg   {channel}: {message} [{self.channels[channel].id}]')  # noqa: E501
             if not self.running:
                 break
 
             if channel is Message.DROP:
+                self.debug(f'*** dropping message: {message}')
                 self.parent.drop(message)
             else:
                 self.debug(f'Sending message to {channel}: {message}')
@@ -423,19 +684,25 @@ class Outbox(pipe.Manifold):
                     try:
                         await self.try_while_running(partial(self.channels[channel].send, message))
                     except ComponentInterrupted:
+                        self.debug(f'*** aborted sending message to {channel}: {message}')
                         break
 
+                    self.debug(f'*** sent to   {channel}: {message} [{self.channels[channel].id}]')  # noqa: E501
                     message.checkout(self)
+                    self.debug(f'*** yielding  {channel}: {message}')
                     yield channel, message
 
+                    self.debug(f'*** yielded   {channel}: {message}')
                 else:
                     raise KeyError(f'Channel "{channel}" does not exist in {self}')
 
         self.debug('Finished sending')
         try:
             for _pipe in self.channels.values():
+                self.debug(f'*** sending {pipe.EOT} to {_pipe}')
                 await self.try_while_running(partial(_pipe.send, pipe.EOT))
         except ComponentInterrupted:
+            self.debug(f'*** aborted sending {pipe.EOT} to {_pipe}')
             pass
 
 
